from fnmatch import fnmatch
import configparser
from logsetup import UserLogger

log = UserLogger(__name__)


def load_config(arguments):
    config = configparser.ConfigParser(allow_no_value=True)
    configpath = arguments['--configfile']

    try:
        with open(configpath) as f:
            config.readfp(f)
    except IOError:
        log.uabort('Config file path not found. ({})', configpath)

    config_dict = {}
    for section in config.sections():
        config_dict[section] = dict((key, True if value is None else value)
                                    for key, value in config.items(section))
    return config_dict


def update_arguments_to_config(arguments, config):
    def key_set(section, key, value):
        s = config.get(section, dict())
        config[section] = s
        s[key] = value
    lookup = {
        "--hostname": ('infinibox', 'host'),
        "--retain": ('global', 'snaps_to_keep'),
        "--basename": ('global', 'snap_basename'),
        "--debug": ('logging', 'debug'),
        "--logfile": ('logging', 'logpath'),
        "--norotation": ('logging', 'rotation'),
        "--password": ('infinibox', 'password'),
        "--quiet": ('logging', 'quiet'),
        "--syslog": ('logging', 'syslog'),
        "--user": ('infinibox', 'username'),
        "--volume": ('infinibox', 'volume'),
        "--purge": ('infinibox', 'purge'),
        "--dryrun": ('global', 'dryrun'),

    }
    for arg_key, (section, key) in lookup.items():
        if arg_key in arguments and arguments[arg_key] != None and arguments[arg_key] != False:
            key_set(section, key, arguments[arg_key])


def validate_config(config):
    from cerberus import Validator

    class MyValidator(Validator):

        def _validate_bool(self, param, field, value):
            if str(value).lower() not in ['true', 'false', 'yes', 'no', 'on', 'off', '1', '0']:
                self._error(field, "Not in bool list")
            else:
                if str(value).lower() in ['true', 'yes', 'on', '1']:
                    self.document[param][field] = True
                else:
                    self.document[param][field] = False

        def _validate_retention(self, param, field, value):
            if not value.isdigit():
                self._error(field, "Not a digit list")
            else:
                retent = int(value)
                self.document[param][field] = retent
                if retent > 254 or retent < 1:
                    self._error(field, "Retention must be between 1 and 254")

        def _validate_parallel(self, param, field, value):
            if not value.isdigit():
                self._error(field, "Not a digit list")
            else:
                parallel = int(value)
                self.document[param][field] = parallel
                if parallel > 20 or parallel < 1:
                    self._error(field, "Parallel must be between 1 and 20")

    schema = {
        'global': {'type': 'dict', 'required': True, 'schema': {
            'snap_basename': {'type': 'string', 'required': True},
            'dryrun': {'bool': 'global'},
            'snaps_to_keep': {'retention': 'global'}}},
        'infinibox': {'type': 'dict', 'required': True, 'schema': {
                      'host': {'type': 'string', 'dependencies': ['password', 'username'], 'required': True},
                      'password': {'type': 'string', 'dependencies': ['host', 'username'], 'required': True},
                      'username': {'type': 'string', 'dependencies': ['password', 'host'], 'required': True},
                      'purge': {'bool': 'infinibox', 'dependencies': ['volume']},
                      'volume': {'type': 'string', 'dependencies': ['host'], 'required': False, 'empty': False}}},
        'logging': {'type': 'dict', 'schema': {
            'logpath': {'type': 'string'},
            'emailuser': {'type': 'string', 'dependencies': ['email', 'emailpassword']},
            'emailpassword': {'type': 'string', 'dependencies': ['email', 'emailuser']},
            'emailsender': {'type': 'string', 'dependencies': ['email'], 'required': True},
            'emailrcpt': {'type': 'string', 'dependencies': ['email'], 'required': True},
            'emailserver': {'type': 'string', 'dependencies': ['email'], 'required': True},
            'emailport': {'type': 'string', 'dependencies': ['email']},
            'quiet': {'bool': 'logging'},
            'rotation': {'bool': 'logging'},
            'debug': {'bool': 'logging'},
            'email': {'bool': 'logging'},
            'syslog': {'bool': 'logging'}}},
        'vcenter': {'type': 'dict', 'dependencies': ['vms', 'noquiesce', 'nomemory', 'datastores'], 'required': False, 'schema': {
            'host': {'type': 'string', 'dependencies': ['password', 'username'], 'required': False},
            'password': {'type': 'string', 'dependencies': ['host', 'username'], 'required': False},
            'parallel': {'parallel': 'vcenter'},
            'username': {'type': 'string', 'dependencies': ['host', 'password'], 'required': False}}},
        'datastores': {'type': 'dict', 'required': False},
        'nomemory': {'type': 'dict', 'required': False},
        'noquiesce': {'type': 'dict', 'required': False},
        'vms': {'type': 'dict', 'required': False},
    }

    v = MyValidator(schema)
    result = v.validate(config)

    if 'infinibox' not in config:
        result = False
    else:
        if 'volume' not in config['infinibox'] and 'vcenter' not in config:
            result = False
            v._error(
                'volume', 'Either [vmware] section or --volume should be configured.')
        elif 'volume' in config['infinibox'] and 'vcenter' in config:
            result = False
            v._error(
                'volume', '[vmware] section and --volume can\'t be configured simultaneously.')

        if 'global' not in config:
            result = False
            v._error(
                'snap_basename', 'snap_basename (--basename or -b) and snaps_to_keep (--retain or -r)/--purge is required.')
        else:
            if 'snaps_to_keep' in config['global'] and 'purge' in config['infinibox']:
                if config['infinibox']['purge']:
                    result = False
                    v._error(
                        'purge', 'Either --purge or --snaps_to_keep should be specified, not both.')
            elif 'snaps_to_keep' not in config['global'] and 'purge' not in config['infinibox']:
                result = False
                v._error(
                    'snaps_to_keep', 'Either --purge or --snaps_to_keep should be specified')

            if 'purge' in config['infinibox']:
                if config['infinibox']['purge']:
                    config['global']['snaps_to_keep'] = 1

    return result, v


def read_datastore_patterns_from_config(config):
    return config['datastores']


def read_vm_patterns_from_config(config):
    return config['vms']


def read_noquiesce_patterns_from_config(config):
    return config['noquiesce']


def read_nomemory_patterns_from_config(config):
    return config['nomemory']


def iter_filtered_datastores(datastores, pattern):
    for datastore in datastores:
        if fnmatch(datastore.name.lower(), pattern.lower()):
            yield datastore


def get_datastores_to_snapshot(datastores, patterns):
    def get_filtered_datastores(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            result.update(iter_filtered_datastores(datastores, pattern))
        return result

    includes = get_filtered_datastores(lambda p: p and p[0] != '!')
    excludes = get_filtered_datastores(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_datastores_to_snapshot(config, datastores):
    patterns = read_datastore_patterns_from_config(config)
    return get_datastores_to_snapshot(datastores, patterns)


def iter_filtered_vms(pattern, datastores, vms):
    datastore_names = [datastore.name for datastore in datastores]

    for vm in vms:
        for datastore in vm.datastore:
            if datastore.name in datastore_names:
                if fnmatch(vm.name.lower(), pattern.lower()):
                    yield vm


def get_vms_to_snapshot(patterns, datastores, vms):
    def get_filtered_vms(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            result.update(iter_filtered_vms(pattern, datastores, vms))
        return result

    includes = get_filtered_vms(lambda p: p and p[0] != '!')
    excludes = get_filtered_vms(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_vms_to_snapshot(config, datastores, vms):
    patterns = read_vm_patterns_from_config(config)
    return get_vms_to_snapshot(patterns, datastores, vms)


def list_of_vms_without_quiesce(config, datastores, vms):
    patterns = read_noquiesce_patterns_from_config(config)
    return get_vms_to_snapshot(patterns, datastores, vms)


def list_of_vms_without_memory(config, datastores, vms):
    patterns = read_nomemory_patterns_from_config(config)
    return get_vms_to_snapshot(patterns, datastores, vms)


def iter_vms(config, datastores, all_vms):
    vms = list_of_vms_to_snapshot(config, datastores, all_vms)
    noquiesce = list_of_vms_without_quiesce(config, datastores, vms)
    nomemory = list_of_vms_without_memory(config, datastores, vms)
    for vm in vms:
        quiesce = vm not in noquiesce
        memory = vm not in nomemory
        yield vm,  quiesce, memory


def is_dryrun(config):
    return 'global' in config and 'dryrun' in config['global'] and config['global']['dryrun']


def is_vmware(config):
    return 'vcenter' in config
