import re
from logsetup import UserLogger

log = UserLogger(__name__)

def filter_irrelevant_snapshots(snapshots, volume_name, basename):
    def pred(snapshot):
        name = snapshot.get_name()
        prefix = r'^{}\.{}\.'.format(re.escape(volume_name), re.escape(basename))
        date = r'\d{4}\.\d{2}\.\d{2}_\d{6}$'  # yyyy.mm.dd_hhmmss
        return bool(re.match(prefix + date, name))
    return [s for s in snapshots if pred(s)]


def make_snapshot_name(volname, basename, time):
    from datetime import datetime
    dt = datetime.fromtimestamp(time)
    return "{}.{}.{}".format(volname, basename, dt.strftime("%Y.%m.%d_%H%M%S"))


def sort_and_filter_snapshots(snapshots, volume_name, basename):
    """:return: sorted list of snapshots where snapshots
    not conforming to the name template are removed"""
    return sorted(filter_irrelevant_snapshots(snapshots, volume_name, basename),
                  key=lambda o: o.get_name())


def trim_number_of_snapshots(snapshots, volume_name, basename, retain, dryrun):
    """We try to delete as many snapshots as needed to keep the number of snapshots less or equal to self.retain."""
    snapshots = sort_and_filter_snapshots(snapshots, volume_name, basename)
    num_to_delete = max(0, len(snapshots) - retain)
    snapshots_to_delete = snapshots[:num_to_delete]
    metadata = list()
    for snap in snapshots_to_delete:
        metadata.append(snap.get_metadata())
        name = None
        try:
            name = snap.get_name()
            if not dryrun:
                snap.delete()
            log.uinfo("Deleted snapshot {}".format(name))
        except:
            log.uwarn("Failed to delete snapshot {}, moving on".format(name), exc_info=True)
    return metadata


def create_latest_snapshot(volume, basename, dryrun):
    """ Create latest snapshot <volname>.<basename>.YYYY.MM.DD_HHMMSS"""
    from time import time
    try:
        name = make_snapshot_name(volume.get_name(), basename, time())
        log.uinfo("Creating snapshot {} for volume {}".format(name, volume.get_name()))
        if dryrun:
            return None
        return volume.create_snapshot(name)
    except:
        log.uwarn("Failed to create latest snapshot", exc_info=True)
        return None
