"""

Usage:
   snaprotator [options]
   snaprotator -h | --help
   snaprotator --version


Options:
   -c configfile --configfile=<configfile>        Config file location
   -H hostname --hostname=<hostname>              DNS name of the infinibox
   -u user --user=<username>                      Administrator's username
   -p password --password=<password>              Administrator's password
   -v volume --volume=<volume>                    Volume to rotate the snaphsots on
   -b basename --basename=<basename>              Basename (daily,weekly, etc...)
   -r retain --retain=<retain>                    Number of snapshots to retain
   -l logfile --logfile=<logfile>                 Use log file instead of stderr
   -h, --help                                     Show this screen
   -D, --debug                                    Debug
   -q, --quiet                                    Suppress all informational messages
   -d, --dryrun                                   Show what will be done
   -P, --purge                                    Purge all snapshots with basename <basename>
   -s, --syslog                                   Log to syslog daemon
   -n, --norotation                               Do NOT rotate logfile, keep writing indefinitely
   -V, --version                                  Show snaprotator version
"""
import signal
import sys
import logbook
from logsetup import UserLogger, setup_logging, setup_mail_logging

from docopt import docopt
from __version__ import __version__ as snaprotator_version
from rotation import (create_latest_snapshot, trim_number_of_snapshots)

import vmware
import config as conf
from infinibox import connect_to_infinibox, infinibox_validate, volume_validate
from infi.pyvmomi_wrapper import Client
from time import sleep
import powertools

log = UserLogger(__name__)

MAX_ATTEMPTS = 3
RETRY_IN = 5
MAX_RETAIN = 250
DEFAULT_PARALLEL = 5


def signal_handler(signal, frame):
    msg = (
        'Execution has been interrupted by Ctrl+C!\n' +
        'Please re-run the tool, to make sure the rotation is in order'
    )
    log.uwarn(msg)
    sys.exit(0)


def connect_to_vcenter(config):
    from requests.exceptions import (ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError)
    from pyVmomi import vim

    hostname = config['vcenter']['host']
    username = config['vcenter']['username']
    password = config['vcenter']['password']

    log.uinfo("Connecting to VCenter '{}'".format(hostname))
    try:
        client = Client(hostname, username=username, password=password)
    except (ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError) as e:
        log.uabort_with_exc_info("VCenter connection error, aborting. {}", str(e))
    except vim.fault.InvalidLogin as e:
        log.uabort_with_exc_info("VCenter login error, aborting. {}", e.msg)
    except SystemExit:
        raise
    except:
        e = sys.exc_info()[0]
        log.uabort_with_exc_info("Failed to connect to VCenter, aborting. Unknown error: {}", str(e))

    return client


def pipeline_vm_tasks(parallel, items, task_factory, task_handler):
    tasks = dict()
    while items or tasks:
        if len(tasks) < parallel and items:
            item = items.pop()
            task = task_factory(item)
            if task is not None:
                tasks[item] = task

        for item, task in tasks.copy().iteritems():
            if task_handler(item, task):
                del tasks[item]
        sleep(0.1)


def do_infinibox_snapshot(infinibox, basename, retain, volume_name, dryrun):
    volume, snapshots = volume_validate(infinibox, volume_name)
    snapshot_obj = create_latest_snapshot(volume, basename, dryrun)
    deleted_snaps_metadata = trim_number_of_snapshots(snapshots, volume_name, basename, retain - 1, dryrun)
    return snapshot_obj, deleted_snaps_metadata


def iter_filtered_datastores_volumes(client, config, infinibox):
    datastores = vmware.get_all_infinidat_datastores(client)
    filtered_datastores = conf.list_of_datastores_to_snapshot(config, datastores)
    infi_datastores = vmware.iter_filtered_datastores_block_devices(client, filtered_datastores)
    for datastore, infinidat_devices in infi_datastores:
        for device in infinidat_devices:
            volume = infinibox.objects.Volume.find(serial=device)
            if volume:
                yield datastore, volume


def get_datastore_infinibox_volumes(client, config, infinibox):
    return list(iter_filtered_datastores_volumes(client, config, infinibox))


def get_datastore_snapshot(infinibox_snapshot, datastores_volumes):
    for datastore, volume_list in datastores_volumes:
        if infinibox_snapshot.get_parent() in volume_list:
            return datastore, infinibox_snapshot


def get_vms_infinisnapid_datastore(datastores_snapshot):
    datastore, infinisnap = datastores_snapshot
    vms = vmware.get_all_vms_on_datastore_with_single_datastore(datastore)
    return vms, infinisnap.get_id(), datastore


def get_vms_in_datastore_within_infinisnapid(client, config, infinibox, infinibox_snapshot):
    datastore_volumes = get_datastore_infinibox_volumes(client, config, infinibox)
    datastore_snapshot = get_datastore_snapshot(infinibox_snapshot, datastore_volumes)

    return get_vms_infinisnapid_datastore(datastore_snapshot)


def parallel_create_vmware_snapshots(client, parallel, vms, basename):
    vmware_snapshots = []

    def create_snap_task_factory(vm_tuple):
        vm, quiesce, memory = vm_tuple
        return vmware.initiate_snapshot_for_vm(client, basename, vm, quiesce=quiesce, memory=memory)

    def create_snap_task_handler(vm_tuple, task):
        try:
            vm = vm_tuple[0]
            state, message = vmware.get_task_status(task)
            if state == 'success':  # success
                snapshot = vmware.wait_for_result(task)
                vmware_snapshots.append((vm, snapshot))
                log.uinfo('Successfully created snapshot for VM {} ({})', vm.name, snapshot)
            elif state == 'error':
                log.uwarn('Snapshot creation failed. VM Name: {}, reason: {}', vm.name, message)
        except:
            log.uwarn('Unexpected error in create snap', exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(parallel, vms, create_snap_task_factory, create_snap_task_handler)
    return vmware_snapshots


def parallel_remove_vmware_snapshots(parallel, vmware_snapshots):
    def remove_snap_task_factory(vm_snapshot_pair):
        return vmware.initiate_remove_snapshot(vm_snapshot_pair)

    def remove_snap_task_handler(vm_snapshot_pair, task):
        try:
            vm, snapshot = vm_snapshot_pair
            state, message = vmware.get_task_status(task)
            if state == 'success':  # success
                log.uinfo('Successfully deleted snapshot {} for VM {}', snapshot, vm.name)
            elif state == 'error':
                log.uwarn('Snapshot deletion failed. VM Name: {}, reason: {}', vm.name, message)
        except:
            log.uwarn('Unexpected error in remove snap', exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(parallel, vmware_snapshots, remove_snap_task_factory, remove_snap_task_handler)


def cleanup_vmware_snapshots(vms, client, basename):
    for vm, quiesce, memory in vms:
        try:
            log.debug('Cleanup VMWare snapshots for host {}', vm.name)
            vmware.cleanup_infinidat_snapshots(client, basename, vm)
        except:
            log.uwarn('Failed to cleanup snapshot {}', str(sys.exc_info()[0]))


def do_infinibox_vmware_snapshots(client, config, parallel, retain, infinibox, basename, dryrun):
    datastores_to_snapshot = get_datastore_infinibox_volumes(client, config, infinibox)
    if not datastores_to_snapshot:
        log.uabort("Can't find datastores to snapshot. Exiting.")

    datastores, infinibox_volumes = zip(*datastores_to_snapshot)
    datastore_names = [datastore.name for datastore in datastores]
    log.uinfo('Datastores to snap: {}', datastore_names)

    all_vms = list(vmware.iter_all_vms(client))
    vms = list(conf.iter_vms(config, datastores, all_vms))

    if not vms:
        log.uwarn('No VMs found. Continuing infinibox snapshots without OS consistency.')

    vms_names = [vm.name for vm, quiesce, memory in vms]
    log.uinfo('VMS to snap: {}', vms_names)

    volume_names = [volume.get_name() for vlist in infinibox_volumes for volume in vlist]

    log.uinfo('Infinibox volumes to snap: {}', volume_names)
    powertools_info = None
    if not dryrun:
        vmware_snapshots = parallel_create_vmware_snapshots(client, parallel, list(vms), basename)
        powertools_info = powertools.get_powertools_info(client)

    for infinibox_volume in volume_names:
        snapshot_name, metadata_list = do_infinibox_snapshot(infinibox, basename, retain, infinibox_volume, dryrun)
        if not dryrun and powertools_info:
            vms_infinisnapid_datastore = get_vms_in_datastore_within_infinisnapid(
                client, config, infinibox, snapshot_name)
            powertools.register_snap_unregister_snaps(powertools_info, vms_infinisnapid_datastore, metadata_list)
        else:
            log.uinfo('No Host Power Tools For VMWare found. Check appliance installed/powered on.')

    if not dryrun:
        parallel_remove_vmware_snapshots(parallel, vmware_snapshots)
        cleanup_vmware_snapshots(vms, client, basename)


def do_purge(infinibox, basename, volume_name, dryrun):
    _, snapshots = volume_validate(infinibox, volume_name)
    deleted_snaps_metadata = trim_number_of_snapshots(snapshots, volume_name, basename, 0, dryrun)
    # fixme add vmware deletion
    return deleted_snaps_metadata


def main():
    signal.signal(signal.SIGINT, signal_handler)
    arguments = docopt(__doc__, version=snaprotator_version)

    config = conf.load_config(arguments) if arguments['--configfile'] else {}
    conf.update_arguments_to_config(arguments, config)
    if not config:
        docopt(__doc__, '-h')

    validate, error = conf.validate_config(config)
    if not validate:
        sys.stderr.write('Configuration file or command line parameters error: {}\n'.format(error.errors))
        raise SystemExit()

    log_handler = setup_logging(config)

    basename = config['global']['snap_basename']
    retain = config['global']['snaps_to_keep']
    if 'purge' in config['infinibox']:
        purge = config['infinibox']['purge']
    else:
        purge = False
    dryrun = conf.is_dryrun(config)
    is_vmware = conf.is_vmware(config)
    mail_handler = setup_mail_logging(config, dryrun)
    with logbook.NullHandler().applicationbound(), mail_handler.applicationbound():
        with log_handler.applicationbound():
            try:
                if dryrun:
                    log.uinfo('DRYRUN')
                infinibox = connect_to_infinibox(config)
                infinibox_validate(infinibox)

                if purge:
                    do_purge(infinibox, basename, config['infinibox']['volume'], dryrun)
                elif is_vmware:
                    client = connect_to_vcenter(config)
                    parallel = config['vcenter']['parallel'] if 'parallel' in config['vcenter'] else DEFAULT_PARALLEL
                    do_infinibox_vmware_snapshots(
                        client, config, parallel, retain, infinibox, basename, dryrun)
                else:  # infinibox only
                    do_infinibox_snapshot(infinibox, basename, retain, config['infinibox']['volume'], dryrun)

                log.uinfo('Done.')
            except SystemExit:
                sys.exit(1)
            except:
                log.error("Internal error occurred, please contact Infinidat support.", exc_info=True, user=True)
                sys.exit(2)
